{-# LANGUAGE GADTs                 #-}
{-# LANGUAGE DataKinds             #-}
{-# LANGUAGE TypeFamilies          #-}
{-# LANGUAGE TypeOperators         #-}
{-# LANGUAGE StandaloneDeriving    #-}
{-# LANGUAGE ScopedTypeVariables   #-}
{-# LANGUAGE NoImplicitPrelude     #-}
{-# LANGUAGE UndecidableInstances  #-}
{-# LANGUAGE TemplateHaskell       #-}
{-# LANGUAGE QuasiQuotes           #-}
{-# LANGUAGE PolyKinds             #-}
{-# LANGUAGE UnicodeSyntax         #-}

-- | Defines natural numbers on both term and type level
module Data.Nat where

import Relation.Equality
import Data.Bool
import Data.Integer
import Data.Function
import Data.Singletons.TH hiding ((:<), (:>), (%:<), Refl, Min, sMin, Max, sMax)
import Data.TypeClass     as T

singletons [d|

  -- | Natural numbers type
  data ℕ ∷ ★ where
    Z ∷ ℕ     -- ^ Zero
    S ∷ ℕ → ℕ -- ^ Successor

  infixl 6 +
  infixl 7 ×
  
  -- | Predecessor of n  
  pred ∷ ℕ → ℕ
  pred Z     = Z
  pred (S n) = n

  -- | Addition
  (+) ∷ ℕ → ℕ → ℕ
  Z     + m = m
  (S n) + m = S (n + m)

  -- | Subtraction
  (-) ∷ ℕ → ℕ → ℕ
  Z     -    _  = Z
  n     -    Z  = n
  (S n) - (S m) = n - m

  -- | Multiplication
  (×) ∷ ℕ → ℕ → ℕ
  Z     × m = Z
  (S n) × m = (n × m) + m

  -- | Exponentiation
  (^) ∷ ℕ → ℕ → ℕ
  n ^ Z     = (S Z)
  n ^ (S m) = n × (n ^ m)

  -- | Ordering relation <
  (<) ∷ ℕ → ℕ → 𝔹
  Z     < Z     = F
  Z     < (S _) = T
  (S _) < Z     = F
  (S n) < (S m) = n < m

  -- | Ordering relation >
  (>) ∷ ℕ → ℕ → 𝔹
  n > m = (¬) (n < m)

  min ∷ ℕ → ℕ → ℕ
  min Z    _ = Z
  min _    Z = Z
  min (S n) (S m) = S (min n m)

  max ∷ ℕ → ℕ → ℕ
  max Z        n  = n
  max n        Z  = n
  max (S n) (S m) = S (max n m)

  fact ∷ ℕ → ℕ
  fact Z     = S Z
  fact (S n) = (S n) × (fact n)
  |]

deriving instance Eq ℕ

deriving instance Show ℕ

instance Num ℕ where
  (+) = (Data.Nat.+)
  (-) = (Data.Nat.-)
  (*) = (Data.Nat.×)
  abs = id
  signum = const 1
  negate = id
  fromInteger = _ℤtoℕ

instance Enum ℕ where
  toEnum   = _ℤtoℕ ∘ T.toInteger
  fromEnum = T.fromInteger ∘ _ℕtoℤ

instance Bounded ℕ where
  minBound = Z
  maxBound = (∞)


-- | Convert a natural into an integer
_ℕtoℤ ∷ ℕ → ℤ
_ℕtoℤ Z     = 0
_ℕtoℤ (S n) = 1 T.+ _ℕtoℤ n


-- | Convert an integer into a natural
_ℤtoℕ ∷ ℤ → ℕ
_ℤtoℕ 0 = Z
_ℤtoℕ n
  | n T.<= 0 = Z
  | n T.>  0 = S (_ℤtoℕ (T.pred n))


-- | Infinity
(∞) ∷ ℕ
(∞) = S (∞)


-- Proofs

-- | Lemma
switch ∷ Sℕ n → Sℕ m → (n :+ S m) :≡ (S n :+ m)
switch SZ     m = Refl
switch (SS n) m = cong SS (switch n m)

-- | Lemma
succ_right ∷ Sℕ n → Sℕ m → S (n :+ m) :≡ (n :+ S m)
succ_right SZ     m = Refl
succ_right (SS n) m = cong SS (succ_right n m)

-- | Lemma
succ_left ∷ Sℕ n → Sℕ m → S (n :+ m) :≡ (S n :+ m)
succ_left n m = succ_right n m ∾ switch n m

-- | Lemma
double ∷ Sℕ (S n) -> n :+ n :≡ S (n :+ Pred n)
double (SS n) = gcastWith (succ_pred (SS n)) (sym (succ_right n (sPred n)))

-- | Relation between successor and addition
succ_plus ∷ Sℕ n → (n :+ S Z) :≡ S n
succ_plus n = switch n SZ ∾ plus_right_id (SS n)

-- | Successor of predecessor
succ_pred ∷ Sℕ (S n) → S (Pred n) :≡ n
succ_pred (SS (SS n)) = Refl

-- | Right identity of addition
plus_right_id ∷ Sℕ n → (n :+ Z) :≡ n
plus_right_id SZ     = Refl
plus_right_id (SS n) = cong SS (plus_right_id n)

-- | Left identity of addition
plus_left_id ∷ Sℕ n → (Z :+ n) :≡ n
plus_left_id n = (plus_commut SZ n) ∾ (plus_right_id n)

-- | Associative property of addition
plus_assoc ∷ Sℕ a → Sℕ b → Sℕ c → (a :+ b) :+ c :≡ a :+ (b :+ c)
plus_assoc SZ     _ _ = Refl
plus_assoc (SS a) b c = cong SS (plus_assoc a b c)

-- | Commutative property of addition
plus_commut ∷ Sℕ n → Sℕ m → (n :+ m) :≡ (m :+ n)
plus_commut n  SZ     = plus_right_id n
plus_commut n  (SS m) =
  sym (succ_right n m) ∾ cong SS (plus_commut n m) ∾ succ_left m n

-- | Left identity of multiplication
cross_left_id ∷ Sℕ n →  (S Z) :× n :≡ n
cross_left_id SZ     = Refl
cross_left_id (SS n) = cong SS (cross_left_id n)

-- | Minimum of itself
min_self ∷ Sℕ n → Min n n :≡ n
min_self SZ     = Refl
min_self (SS n) = cong SS (min_self n)

-- | Maximum of itself
max_self ∷ Sℕ n → Max n n :≡ n
max_self SZ     = Refl
max_self (SS n) = cong SS (max_self n)

